@@ -2,8 +2,8 @@
  * AT-SPI - Assistive Technology Service Provider Interface
  * (Gnome Accessibility Project; http://developer.gnome.org/projects/gap)
  *
- * Copyright 2001, 2002 Sun Microsystems Inc.,
- * Copyright 2001, 2002 Ximian, Inc.
+ * Copyright 2001, 2002, 2003 Sun Microsystems Inc.,
+ * Copyright 2001, 2002, 2003 Ximian, Inc.
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Library General Public
@@ -29,11 +29,19 @@ static void traverse_accessible_tree (Accessible *accessible);
 
 static void report_event  (const AccessibleEvent *event, void *user_data);
 static void report_detail_event  (const AccessibleEvent *event, void *user_data);
+static void report_detail1_event  (const AccessibleEvent *event, void *user_data);
+static void report_text_event  (const AccessibleEvent *event, void *user_data);
+static void report_children_changed_event (const AccessibleEvent *event, void *user_data);
 static void timing_test_event (const AccessibleEvent *event, void *user_data);
+static SPIBoolean report_mouse_event  (const AccessibleDeviceEvent *event, void *user_data);
 
 static AccessibleEventListener *generic_listener;
 static AccessibleEventListener *specific_listener;
+static AccessibleEventListener *detail1_listener;
 static AccessibleEventListener *test_listener;
+static AccessibleEventListener *text_listener;
+static AccessibleEventListener *children_changed_listener;
+static AccessibleDeviceListener *mouse_device_listener;
 static gint n_elements_traversed = 0;
 static GTimer *timer;
 
@@ -78,14 +86,24 @@ main (int argc, char **argv)
       }
   }
 
+  fprintf (stderr, "RUNNING\n");
+
   SPI_init ();
 
   generic_listener = SPI_createAccessibleEventListener (
 	  report_event, NULL); 
   specific_listener = SPI_createAccessibleEventListener (
 	  report_detail_event, NULL); 
+  text_listener = SPI_createAccessibleEventListener (
+	  report_text_event, NULL);
+  children_changed_listener = SPI_createAccessibleEventListener (
+	  report_children_changed_event, NULL);
   test_listener = SPI_createAccessibleEventListener (
 	  timing_test_event, NULL);
+  mouse_device_listener = SPI_createAccessibleDeviceListener (
+          report_mouse_event, NULL);
+  detail1_listener = SPI_createAccessibleEventListener (
+	  report_detail1_event, NULL); 
 
   SPI_registerGlobalEventListener (generic_listener,
 				   "focus:");
@@ -97,6 +115,11 @@ main (int argc, char **argv)
       SPI_registerGlobalEventListener (specific_listener,
 				       "mouse:abs");
   }
+  SPI_registerDeviceEventListener (mouse_device_listener, 
+				   SPI_BUTTON_PRESSED | SPI_BUTTON_RELEASED,
+				   NULL);
+  SPI_registerGlobalEventListener (specific_listener,
+				   "keyboard:modifiers");
   SPI_registerGlobalEventListener (generic_listener,
 				   "object:property-change");
 /*  SPI_registerGlobalEventListener (specific_listener,
@@ -107,10 +130,10 @@ main (int argc, char **argv)
     "object:state-changed:focused"); */
   SPI_registerGlobalEventListener (generic_listener,
 				   "object:selection-changed"); 
-  SPI_registerGlobalEventListener (generic_listener,
+  SPI_registerGlobalEventListener (children_changed_listener,
 				   "object:children-changed"); 
-/*  SPI_registerGlobalEventListener (specific_listener,
-    "object:children-changed:add"); */
+  SPI_registerGlobalEventListener (generic_listener,
+				   "object:active-descendant"); 
   SPI_registerGlobalEventListener (generic_listener,
 				   "object:visible-data-changed"); 
   SPI_registerGlobalEventListener (generic_listener,
@@ -118,7 +141,7 @@ main (int argc, char **argv)
 
   SPI_registerGlobalEventListener (generic_listener,
 				   "object:text-caret-moved"); 
-  SPI_registerGlobalEventListener (generic_listener,
+  SPI_registerGlobalEventListener (text_listener,
 				   "object:text-changed"); 
   SPI_registerGlobalEventListener (generic_listener,
 				   "object:column-inserted"); 
@@ -134,6 +157,8 @@ main (int argc, char **argv)
 				   "object:row-deleted"); 
   SPI_registerGlobalEventListener (generic_listener,
 				   "object:model-changed"); 
+  SPI_registerGlobalEventListener (detail1_listener,
+				   "object:link-selected"); 
   SPI_registerGlobalEventListener (generic_listener,
 				   "window:minimize");
   SPI_registerGlobalEventListener (generic_listener,
@@ -251,6 +276,55 @@ report_detail_event (const AccessibleEvent *event, void *user_data)
   if (s) SPI_freeString (s);
 }
 
+void
+report_detail1_event (const AccessibleEvent *event, void *user_data)
+{
+  char *s = Accessible_getName (event->source);
+  fprintf (stderr, "(detail) %s %s %d\n", event->type, s,
+	   event->detail1);
+  if (s) SPI_freeString (s);
+}
+
+void
+report_text_event (const AccessibleEvent *event, void *user_data)
+{
+  char *s = Accessible_getName (event->source);
+  fprintf (stderr, "(detail) %s %s %d %d\n", event->type, s,
+	   event->detail1, event->detail2);
+  SPI_freeString (s);
+  s = AccessibleTextChangedEvent_getChangeString (event);
+  fprintf (stderr, "context string %s\n", (s) ? s : "<nil>");
+  SPI_freeString (s);
+}
+
+void
+report_children_changed_event (const AccessibleEvent *event, void *user_data)
+{
+  char *s = Accessible_getName (event->source);
+  char *s1;
+  Accessible *ao;
+
+  ao = AccessibleChildChangedEvent_getChildAccessible (event);
+  s1 = Accessible_getName (ao);
+  fprintf (stderr, "(detail) %s parent: %s child: %s %d %d\n", event->type, 
+           s ? s : "<null>", s1 ? s1 : "<null>",
+	   event->detail1, event->detail2);
+  SPI_freeString (s);
+  SPI_freeString (s1);
+  Accessible_unref (ao);
+}
+
+SPIBoolean
+report_mouse_event (const AccessibleDeviceEvent *event, void *user_data)
+{
+  fprintf (stderr, "mouse event %ld %d %x %x\n", 
+	   event->keyID, 
+	   (int) event->keycode,
+	   (unsigned) event->type,
+	   (unsigned) event->modifiers);
+  return FALSE;
+}
+
 void
 timing_test_event (const AccessibleEvent *event, void *user_data)
 {
